#include "parser.h"
#include "helpers.h"

Node parse_token(Compiler* self, Token token) {
    switch(token.type) {
    case TOKEN_TYPEIDENTIFIER:
        return parse_type(self, token.value);
    case TOKEN_IDENTIFIER:
        return parse_identifier(self, token.value);
    case TOKEN_NUMBER:
        return parse_number(self, token.value);
    default:
        syntax_error("unexpected token \"%s\"", token.value);
    }
}

Nodes parse_until(Compiler* self, TokenType end_token) {
    Nodes collected = { 0 };

    Token token;
    while(next(self, &token)) {
        if(token.type == end_token) {
            break;
        }

        Nodes_push(&collected, parse_token(self, token));
    }

    return collected;
}

Node parse_type(Compiler* self, const char* type) {
    printf("parse type %s\n", type);
    Token token;
    next(self, &token);

    switch(token.type) {
    case TOKEN_IDENTIFIER:
        return parse_type_ident_pair(self, type, token.value);
        break;
    default:
        syntax_error("unexpected token \"%s\" after type \"%s\"", token.value, type);
    }
}

const char* resolve_type(Compiler* self, Node node) {
    switch(node.type) {
    case NODE_ARG_DECL:
        return node.arg_decl.type;
    default:
        syntax_error("unexpected node fed to resolve_type");
    }
}

Node parse_identifier(Compiler* self, const char* name) {
    if(strcmp(name, "return") == 0) {
        Token token;
        next(self, &token);
        Node* node = NULL;

        if(token.type != TOKEN_SEMICOLON) {
            node = malloc(sizeof(Node));
            Node parsed = parse_token(self, token);
            memcpy(node, &parsed, sizeof(Node));
            parse_semicolon(self);
        }
        return (Node){ .type = NODE_RETURN, .ret = { .value = node } };
    }

    const NodeMap_value* symbol = NodeMap_get(&self->symbols, name);
    if(symbol == NULL) {
        syntax_error("undefined symbol \"%s\"", name);
    }

    switch(symbol->second.type) {
    case NODE_FUNCTION_DECL:
    case NODE_FUNCTION_IMPL:
        return parse_function_call(self, name);
    case NODE_ARG_DECL:
        return (Node){
            .type = NODE_REFERENCE,
            .reference = { .type = resolve_type(self, symbol->second), .name = name },
        };
    default:
        syntax_error("unexpected symbol type");
    }
}

Node parse_number(Compiler* self, const char* value) {
    return (Node){
        .type = NODE_NUMBER,
        .number = atoi(value),
    };
}

Node parse_type_ident_pair(Compiler* self, const char* type, const char* name) {
    printf("parse type ident pair %s %s\n", type, name);
    Token token;
    next(self, &token);

    switch(token.type) {
    case TOKEN_OPAREN:
        return parse_function(self, type, name);
    default:
        syntax_error("unexpected token \"%s\" after \"%s %s\"", token.value, type, name);
    }
}

Node parse_function(Compiler* self, const char* type, const char* name) {
    printf("parse function %s %s\n", type, name);
    Nodes args = { 0 };
    while(true) {
        Token arg_type;
        next(self, &arg_type);
        if(arg_type.type == TOKEN_CPAREN) {
            break;
        }
        const cmap_str_value* type = cmap_str_get(&self->types, arg_type.value);

        Token arg_name;
        next(self, &arg_name);

        Node arg_decl = (Node){
            .type = NODE_ARG_DECL,
            .arg_decl = { .type = cstr_str(&type->second), .name = arg_name.value },
        };
        Nodes_push(&args, arg_decl);
        NodeMap_insert(&self->symbols, cstr_from(arg_name.value), arg_decl);

        Token token;
        next(self, &token);
        switch(token.type) {
        case TOKEN_COMMA:
            continue;
        case TOKEN_CPAREN:
            break;
        default:
            syntax_error("expected comma or closing parenthesis, found %s", token.value);
        }

        break;  // only reached if didnt continue or error
    }

    const cmap_str_value* real_type = cmap_str_get(&self->types, type);
    const char* real_type_cstr = cstr_str(&real_type->second);

    Token token;
    Node function_node;
    if(next(self, &token)) {
        switch(token.type) {
        case TOKEN_SEMICOLON:
            function_node = parse_function_decl(self, real_type_cstr, name, args);
            break;
        case TOKEN_OBRACE:
            function_node = parse_function_impl(self, real_type_cstr, name, args);
            break;
        default:
            syntax_error("expected semicolon or opening brace found, %s", token.value);
        }
    } else {
        syntax_error("expected token, found eof");
    }

    printf("inserting %s into symbols\n", name);
    NodeMap_insert(&self->symbols, cstr_from(name), function_node);
    return function_node;
}

Node parse_function_decl(Compiler* self, const char* type, const char* name, Nodes args) {
    return (Node){
        .type = NODE_FUNCTION_DECL,
        .function_decl = { .type = type, .name = name, .args = args },
    };
}

Node parse_function_impl(Compiler* self, const char* type, const char* name, Nodes args) {
    Nodes body = parse_until(self, TOKEN_CBRACE);

    return (Node){
        .type = NODE_FUNCTION_IMPL,
        .function_impl = { .type = type, .name = name, .args = args, .body = body },
    };
}

Node parse_function_call(Compiler* self, const char* name) {
    printf("parse function call %s\n", name);
    Token token;
    next(self, &token);
    switch(token.type) {
    case TOKEN_OPAREN:
        break;
    default:
        syntax_error("unexpected token \"%s\" after function name", token.value);
    }

    Nodes args = parse_until(self, TOKEN_CPAREN);

    parse_semicolon(self);
    return (Node){
        .type = NODE_FUNCTION_CALL,
        .function_call = { .name = name, .args = args },
    };
}

void parse_semicolon(Compiler* self) {
    Token token;
    next(self, &token);

    switch(token.type) {
    case TOKEN_SEMICOLON:
        return;
    default:
        syntax_error("expected semicolon, found \"%s\"", token.value);
    }
}
